local math  = math
local table = table

PartyVitals = UI:Create( "BASE", nil, function( main )
    main:SetMouseVisible( false )
    main:SetColor( 200, 200, 200, 0 )
    main:SetSize( UI.SCREEN_W * 2, UI.SCREEN_H * 2 )
    main:SetPos( Group.Settings.Pos.x - UI.SCREEN_W, Group.Settings.Pos.y - UI.SCREEN_H )
    main:SetWantsKeyEvents( true )
    main:EnableDragBase()

    function main:MouseDown( mouse )
        if self.shift_status then
            if Group.enable_layout_preview then
                for i = 1, self.preview_count do
                    if self.preview[ i ]:IsVisible() and self.preview[ i ]:IsHovering() then
                        local x, y   = self:GetPos()
                        local mx, my = Turbine.UI.Display.GetMousePosition()
    
                        self.start_x = mx - x
                        self.start_y = my - y

                        self.isDragging = true
    
                        break
                    end
                end
            else
                for i = 1, #self.Drag do
                    local v = self.Drag[ i ]
    
                    if v:IsVisible() and v.draggable and v:IsHovering() then
                        local pos_x, pos_y = v:GetPos()
    
                        self.drag_offset_x = mouse.X - pos_x
                        self.drag_offset_y = mouse.Y - pos_y
    
                        self.isDragging = true
    
                        self.drag_origin = v
    
                        v.origin_x = pos_x
                        v.origin_y = pos_y
    
                        v.isDragging = true
    
                        v:SetZOrder( 10 )
    
                        self:OnDragStart( v )
    
                        break
                    end
                end
            end
        end
    end

    function main:MouseUp( mouse )
        if self.isDragging then
            self.isDragging = false

            if Group.enable_layout_preview then
                local x, y = self:GetPos()

                Group.Settings.Pos.x = x + UI.SCREEN_W
                Group.Settings.Pos.y = y + UI.SCREEN_H

                Turbine.PluginData.Save( Turbine.DataScope.Character, "MiniRaid_Settings", Group.Settings )

                self:LayoutPosChanged( x + UI.SCREEN_W, y + UI.SCREEN_H )

            else
                local target = nil

                for i = 1, #self.Drag do
                    local v = self.Drag[ i ]

                    if not v.isDragging and v:IsVisible() and v:IsHovering() then
                        target = v

                        break
                    end
                end

                self.drag_origin.isDragging = false

                self.drag_origin:SetZOrder( 1 )

                self.drag_origin:SetPos( self.drag_origin.origin_x, self.drag_origin.origin_y )

                self:OnDragEnd( self.drag_origin, target )
            end
        end
    end

    function main:MouseMove( mouse )
        if self.isDragging then
            if Group.enable_layout_preview then
                local mx, my = Turbine.UI.Display.GetMousePosition()

                local x = mx - self.start_x
                local y = my - self.start_y

                local w, h = self:GetSize()

                self:SetPos( x, y )
            else
                self.drag_origin:SetPos( mouse.X - self.drag_offset_x, mouse.Y - self.drag_offset_y )
            end
        end
    end

    function main:OnShiftChanged( shift )
        if shift then
            for i = 1, #self.Drag do
                self.Drag[ i ]:SetMouseVisible( false )
                self.Drag[ i ].entity:SetMouseVisible( false )
            end
        else
            for i = 1, #self.Drag do
                self.Drag[ i ]:SetMouseVisible( true )
                self.Drag[ i ].entity:SetMouseVisible( true )
            end
        end
    end

    function main:OnDragEnd( origin, target )
        if not target then return end

        local origin_slot = Group.PartyName[ origin:GetName() ]
        local target_slot = Group.PartyName[ target:GetName() ]

        Group.Party[ origin_slot ] = target
        Group.Party[ target_slot ] = origin

        Group.PartyName[ origin:GetName() ] = target_slot
        Group.PartyName[ target:GetName() ] = origin_slot

        Group:RefreshLayout()
    end

    main.ui_toggle    = false
    main.save_visible = main:IsVisible()

    function main:KeyDown( key )
        if key.Action == Turbine.UI.Lotro.Action.Escape then
            if self.ui_toggle then
                self.ui_toggle = false

                self:SetVisible( self.save_visible )
            end
        elseif key.Action == Turbine.UI.Lotro.Action.UI_Toggle then
            self.disable_changed = true

            if self.ui_toggle then
                if not Group.hidden then
                    self:SetVisible( self.save_visible )
                end
            else
                self:SetVisible( false )
            end

            self.disable_changed = false

            self.ui_toggle = not self.ui_toggle
        end
    end

    function main:VisibleChanged()
        if not self.disable_changed then
            self.save_visible = self:IsVisible()
        end
    end

    function main:LayoutPosChanged( x, y )
        --Override
    end

    main.preview = {}
    main.preview_count = 12

    local settings = Group.Settings

    for i = 1, 24 do
        main.preview[ i ] = UI:Create( "BOX", main, function( ui )
            ui:SetSize( settings.vitals_x, settings.vitals_y )
            ui:SetMouseVisible( false )
            ui:SetVisible( false )

            ui.text = UI:Create( "TEXT", ui, function( ui )
                ui:SetSize( settings.vitals_x, settings.vitals_y )
                ui:SetFont( Turbine.UI.Lotro.Font.FixedSys15 )
                ui:SetTextAlignment( 5 )
                ui:SetFontStyle( Turbine.UI.FontStyle.Outline )
                ui:SetText( i )
                ui:SetMouseVisible( false )
            end )

            function ui:SizeChanged()
                local w, h = self:GetSize()

                self.text:SetSize( w, h )
            end
        end )
    end

    function main:DefinePreviewLayout( count )
        local settings = Group.Settings

        if count <= 3 then
            self.grid_x = settings.Layout[ 1 ].x
            self.grid_y = settings.Layout[ 1 ].y
        elseif count <= 6 then
            self.grid_x = settings.Layout[ 2 ].x
            self.grid_y = settings.Layout[ 2 ].y
        elseif count <= 12 then
            self.grid_x = settings.Layout[ 3 ].x
            self.grid_y = settings.Layout[ 3 ].y
        elseif count <= 18 then
            self.grid_x = settings.Layout[ 4 ].x
            self.grid_y = settings.Layout[ 4 ].y
        elseif count <= 24 then
            self.grid_x = settings.Layout[ 5 ].x
            self.grid_y = settings.Layout[ 5 ].y
        end
    end

    function main:RefreshPreview()
        local slot = 1
    
        local w, h = self:GetSize()

        local settings = Group.Settings
    
        for y = 1, self.grid_y do
            for x = 1, self.grid_x do
                local count    = self.preview_count - self.grid_x * ( y - 1 )
                local row_size = self.grid_x
    
                if y == self.grid_y and self.grid_x < count then
                    row_size = count - self.grid_x
                elseif ( y == self.grid_y and self.grid_x > count ) or self.grid_x > count then
                    row_size = count
                end
    
                local pos_x
                local pos_y
    
                if settings.horizontal_layout_direction == 1 then --Left
                    pos_x = w / 2 + settings.vitals_x * ( x - 1 ) + settings.vitals_external_spacing * ( x - 1 )
                elseif settings.horizontal_layout_direction == 2 then --Center
                    pos_x = w / 2 + settings.vitals_x * ( x - 1 ) + settings.vitals_external_spacing * ( x - 1 ) - ( settings.vitals_x * row_size + settings.vitals_external_spacing * ( row_size - 1 ) ) / 2
                else --Right
                    pos_x = w / 2 - settings.vitals_x * x - settings.vitals_external_spacing * ( x - 1 )
                end
    
                if settings.vertical_layout_direction == 1 then --Top
                    pos_y = h / 2 + settings.vitals_y * ( y - 1 ) + settings.vitals_external_spacing * ( y - 1 )
                else --Bottom
                    pos_y = h / 2 - settings.vitals_y * y - settings.vitals_external_spacing * ( y - 1 )
                end
    
                self.preview[ slot ]:SetPos( pos_x, pos_y )
    
                slot = slot + 1
    
                if slot > self.preview_count then return end
            end
        end
    end
end )

UI:CreateTemplate( "BASE", "FRAME", function( template )
    function template:Init( main )
        main:SetColor( Group.Settings.Color.Background )
        main:SetSize( Group.Settings.vitals_x, Group.Settings.vitals_y )
        main:SetPos( UI.SCREEN_W / 2, UI.SCREEN_H / 2 )

        main:SetWantsUpdates( true )

        main.callback        = {}
        main.callback_effect = {}

        main.effect       = {}
        main.effect_list  = {}
        main.effect_count = 20

        main.Pos = {
            [ 1 ] = function( ui, x, y, w, h ) 
                local settings = Group.Settings
                
                if settings.enable_classic_power_bar then
                    ui:SetPos( x, y )
                else
                    ui:SetPos( settings.power_size + x, y ) 
                end
            end, --Top left
            
            [ 2 ] = function( ui, x, y, w, h ) 
                local settings = Group.Settings
                
                if settings.enable_classic_power_bar then
                    ui:SetPos( settings.vitals_x - x - w, y )
                else
                    ui:SetPos( settings.power_size + ( settings.vitals_x - settings.power_size ) - x - w, y ) 
                end
            end, --Top Right
            
            [ 3 ] = function( ui, x, y, w, h ) 
                local settings = Group.Settings
                 
                if settings.enable_classic_power_bar then
                    ui:SetPos( settings.vitals_x - x - w, settings.vitals_y - y - h - settings.power_size )
                else
                    ui:SetPos( settings.power_size + ( settings.vitals_x - settings.power_size ) - x - w, settings.vitals_y - y - h )
                end
            end, --Bottom Right
            
            [ 4 ] = function( ui, x, y, w, h ) 
                local settings = Group.Settings
                
                if settings.enable_classic_power_bar then
                    ui:SetPos( x, settings.vitals_y - y - h - settings.power_size )
                else
                    ui:SetPos( settings.power_size + x, settings.vitals_y - y - h )
                end
            end, --Bottom Left
            
            [ 5 ] = function( ui, x, y, w, h )
                local settings = Group.Settings

                if settings.enable_classic_power_bar then
                    ui:SetPos( settings.vitals_x / 2 - w / 2 + x, ( settings.vitals_y - settings.power_size ) / 2 - h / 2 + y ) 
                else
                    ui:SetPos( settings.power_size + ( settings.vitals_x - settings.power_size ) / 2 - w / 2 + x, settings.vitals_y / 2 - h / 2 + y ) 
                end
            end --Center 
        }

        main.EffectPos = {
            [ 1 ] = function( x, y, w, h, ex, ey )
                local settings = Group.Settings

                if settings.enable_classic_power_bar then
                    local offset_x = settings.effect_size * ( x - 1 ) + settings.effect_spacing * ( x - 1 ) + ex
                    local offset_y = settings.effect_size * ( y - 1 ) + settings.effect_spacing * ( y - 1 ) + ey

                    return offset_x, offset_y
                else
                    local offset_x = settings.effect_size * ( x - 1 ) + settings.effect_spacing * ( x - 1 ) + ex + settings.power_size
                    local offset_y = settings.effect_size * ( y - 1 ) + settings.effect_spacing * ( y - 1 ) + ey

                    return offset_x, offset_y
                end
            end, --Top Left

            [ 2 ] = function( x, y, w, h, ex, ey ) 
                local settings = Group.Settings

                if settings.enable_classic_power_bar then
                    local offset_x = w - settings.effect_size * x - settings.effect_spacing * ( x - 1 ) - ex
                    local offset_y = settings.effect_size * ( y - 1 ) + settings.effect_spacing * ( y - 1 ) + ey

                    return offset_x, offset_y
                else
                    local offset_x = w - settings.effect_size * x - settings.effect_spacing * ( x - 1 ) - ex + settings.power_size
                    local offset_y = settings.effect_size * ( y - 1 ) + settings.effect_spacing * ( y - 1 ) + ey

                    return offset_x, offset_y
                end
            end, --Top Right

            [ 3 ] = function( x, y, w, h, ex, ey ) 
                local settings = Group.Settings

                if settings.enable_classic_power_bar then
                    local offset_x = w - settings.effect_size * x - settings.effect_spacing * ( x - 1 ) - ex
                    local offset_y = h - settings.effect_size * y - settings.effect_spacing * ( y - 1 ) - ey

                    return offset_x, offset_y
                else
                    local offset_x = w - settings.effect_size * x - settings.effect_spacing * ( x - 1 ) - ex + settings.power_size
                    local offset_y = h - settings.effect_size * y - settings.effect_spacing * ( y - 1 ) - ey

                    return offset_x, offset_y
                end
            end, --Bottom Right

            [ 4 ] = function( x, y, w, h, ex, ey ) 
                local settings = Group.Settings

                if settings.enable_classic_power_bar then
                    local offset_x = settings.effect_size * ( x - 1 ) + settings.effect_spacing * ( x - 1 ) + ex
                    local offset_y = h - settings.effect_size * y - settings.effect_spacing * ( y - 1 ) - ey

                    return offset_x, offset_y
                else
                    local offset_x = settings.effect_size * ( x - 1 ) + settings.effect_spacing * ( x - 1 ) + ex + settings.power_size
                    local offset_y = h - settings.effect_size * y - settings.effect_spacing * ( y - 1 ) - ey

                    return offset_x, offset_y
                end
            end --Bottom Left
        }

        main.bubble = UI:Create( "BOX", main, function( ui )
            if Group.Settings.enable_classic_power_bar then
                ui:SetSize( 0, Group.Settings.vitals_y - Group.Settings.power_size )
            else
                ui:SetPos( Group.Settings.power_size, 0 )
                ui:SetSize( 0, Group.Settings.vitals_y )
            end

            ui:SetColor( Group.Settings.Color.Bubble )
            ui:SetBackColorBlendMode( Turbine.UI.BlendMode.AlphaBlend )
            ui:SetMouseVisible( false )
        end )
    
        main.morale = UI:Create( "BOX", main, function( ui, parent )
            if Group.Settings.enable_classic_power_bar then
                ui:SetSize( Group.Settings.vitals_x, Group.Settings.vitals_y - Group.Settings.power_size )
            else
                ui:SetPos( Group.Settings.power_size, 0 )
                ui:SetSize( Group.Settings.vitals_x - Group.Settings.power_size, Group.Settings.vitals_y )
            end

            ui:SetColor( Group.Settings.Color.Morale )
            ui:SetBackColorBlendMode( Turbine.UI.BlendMode.AlphaBlend )
            ui:SetMouseVisible( false )
        end )

        main.power = UI:Create( "BOX", main, function( ui, parent )
            if Group.Settings.enable_classic_power_bar then
                ui:SetPos( 0, Group.Settings.vitals_y - Group.Settings.power_size )
                ui:SetSize( Group.Settings.vitals_x, Group.Settings.power_size )
            else
                ui:SetSize( Group.Settings.power_size, Group.Settings.vitals_y )
            end

            ui:SetColor( Group.Settings.Color.Power )
            ui:SetBackColorBlendMode( Turbine.UI.BlendMode.AlphaBlend )
            ui:SetMouseVisible( false )
        end )

        main.power.text = UI:Create( "TEXT", main, function( ui )
            ui:SetMultiline( false )
            ui:SetSize( 256, Group.Font[ Group.Settings.power_font_type ].size[ Group.Settings.power_font_size ] )
            ui:SetFont( Group:GetFont( Group.Settings.power_font_type, Group.Settings.power_font_size ) )
            ui:SetVisible( Group.Settings.enable_power_text )
            ui:SetMouseVisible( false )

            if Group.Settings.Frame.Power_Text.anchor == 1 or Group.Settings.Frame.Power_Text.anchor == 4 then
                ui:SetTextAlignment( 4 )
            elseif Group.Settings.Frame.Power_Text.anchor == 5 then
                ui:SetTextAlignment( 5 )
            else
                ui:SetTextAlignment( 6 )
            end

            if Group.Settings.power_font_outline then
                ui:SetFontStyle( Turbine.UI.FontStyle.Outline )
            else
                ui:SetFontStyle( Turbine.UI.FontStyle.None )
            end

            ui:SetForeColor( Group.Settings.Color.Power_Text )

            ui.format = {
                [ 1 ] = function( current, max, percent ) --%
                    ui:SetText( math.round( percent ) .. "%" ) 
                end,
                [ 2 ] = function( current, max, percent ) --Current
                    ui:SetText( math.format( math.round( current ) ) ) 
                end,
                [ 3 ] = function( current, max, percent ) --Current + %
                    ui:SetText( math.format( math.round( current ) ) .. " " .. math.round( percent ) .. "%" ) 
                end,
                [ 4 ] = function( current, max, percent ) --Current / Max
                    ui:SetText( math.format( math.round( current ) ) .. "/" .. math.format( math.round( max ) ) ) 
                end,
                [ 5 ] = function( current, max, percent ) --Current / Max + %
                    ui:SetText( math.format( math.round( current ) ) .. "/" .. math.format( math.round( max ) ) .. " " .. math.round( percent ) .. "%" )
                end,
                [ 6 ] = function( current, max, percent ) --Current ( K )
                    ui:SetText( math.formatK( current ) )
                end,
                [ 7 ] = function( current, max, percent ) --Current + % ( K )
                    ui:SetText( math.formatK( current ) .. " " .. math.round( percent ) .. "%" )
                end,
                [ 8 ] = function( current, max, percent ) --Current / Max ( K )
                    ui:SetText( math.formatK( current ) .. "/" .. math.formatK( max ) )
                end,
                [ 9 ] = function( current, max, percent ) --Current / Max + % ( K )
                    ui:SetText( math.formatK( current ) .. "/" .. math.formatK( max ) .. " " .. math.round( percent ) .. "%" )
                end
            }

            ui.Pos = {
                [ 1 ] = function( ui, x, y, w, h ) 
                    local settings = Group.Settings
                    
                    if settings.enable_classic_power_bar then
                        ui:SetPos( x, settings.vitals_y - settings.power_size + y )
                    else
                        ui:SetPos( x, y ) 
                    end
                end, --Top left
                
                [ 2 ] = function( ui, x, y, w, h ) 
                    local settings = Group.Settings
                    
                    if settings.enable_classic_power_bar then
                        ui:SetPos( settings.vitals_x - x - w, settings.vitals_y - settings.power_size + y )
                    else
                        ui:SetPos( settings.power_size - x - w, y ) 
                    end
                end, --Top Right
                
                [ 3 ] = function( ui, x, y, w, h ) 
                    local settings = Group.Settings
                     
                    if settings.enable_classic_power_bar then
                        ui:SetPos( settings.vitals_x - x - w, settings.vitals_y - y - h )
                    else
                        ui:SetPos( settings.power_size - x - w, settings.vitals_y - y - h )
                    end
                end, --Bottom Right
                
                [ 4 ] = function( ui, x, y, w, h ) 
                    local settings = Group.Settings
                    
                    if settings.enable_classic_power_bar then
                        ui:SetPos( x, settings.vitals_y - y - h )
                    else
                        ui:SetPos( x, settings.vitals_y - y - h )
                    end
                end, --Bottom Left
                
                [ 5 ] = function( ui, x, y, w, h )
                    local settings = Group.Settings
    
                    if settings.enable_classic_power_bar then
                        ui:SetPos( settings.vitals_x / 2 - w / 2 + x, settings.vitals_y - settings.power_size / 2 - h / 2 + y ) 
                    else
                        ui:SetPos( settings.power_size / 2 - w / 2 + x, settings.vitals_y / 2 - h / 2 + y ) 
                    end
                end --Center 
            }

            ui.Pos[ Group.Settings.Frame.Power_Text.anchor ]( ui, Group.Settings.Frame.Power_Text.x, Group.Settings.Frame.Power_Text.y, ui:GetSize() )

            function ui:SizeChanged()
                local settings = Group.Settings

                self.Pos[ settings.Frame.Power_Text.anchor ]( self, settings.Frame.Power_Text.x, settings.Frame.Power_Text.y, self:GetSize() )
            end
        end )

        main.dead = UI:Create( "BOX", main, function( ui, parent )
            ui:SetSize( Group.Settings.vitals_x, Group.Settings.vitals_y )
            ui:SetColor( Group.Settings.Color.Dead )
            ui:SetBackColorBlendMode( Turbine.UI.BlendMode.AlphaBlend )
            ui:SetMouseVisible( false )
            ui:SetVisible( false )
        end )

        main.morale.text = UI:Create( "TEXT", main, function( ui )
            ui:SetMultiline( false )
            ui:SetSize( 256, Group.Font[ Group.Settings.morale_font_type ].size[ Group.Settings.morale_font_size ] )
            ui:SetFont( Group:GetFont( Group.Settings.morale_font_type, Group.Settings.morale_font_size ) )
            ui:SetVisible( Group.Settings.enable_morale_text )
            ui:SetMouseVisible( false )

            if Group.Settings.Frame.Morale_Text.anchor == 1 or Group.Settings.Frame.Morale_Text.anchor == 4 then
                ui:SetTextAlignment( 4 )
            elseif Group.Settings.Frame.Morale_Text.anchor == 5 then
                ui:SetTextAlignment( 5 )
            else
                ui:SetTextAlignment( 6 )
            end

            if Group.Settings.morale_font_outline then
                ui:SetFontStyle( Turbine.UI.FontStyle.Outline )
            else
                ui:SetFontStyle( Turbine.UI.FontStyle.None )
            end

            ui:SetForeColor( Group.Settings.Color.Morale_Text )

            ui.format = {
                [ 1 ] = function( current, max, percent ) --%
                    ui:SetText( math.round( percent ) .. "%" ) 
                end,
                [ 2 ] = function( current, max, percent ) --Current
                    ui:SetText( math.format( math.round( current ) ) ) 
                end,
                [ 3 ] = function( current, max, percent ) --Current + %
                    ui:SetText( math.format( math.round( current ) ) .. " " .. math.round( percent ) .. "%" ) 
                end,
                [ 4 ] = function( current, max, percent ) --Current / Max
                    ui:SetText( math.format( math.round( current ) ) .. "/" .. math.format( math.round( max ) ) ) 
                end,
                [ 5 ] = function( current, max, percent ) --Current / Max + %
                    ui:SetText( math.format( math.round( current ) ) .. "/" .. math.format( math.round( max ) ) .. " " .. math.round( percent ) .. "%" )
                end,
                [ 6 ] = function( current, max, percent ) --Current ( K )
                    ui:SetText( math.formatK( current ) )
                end,
                [ 7 ] = function( current, max, percent ) --Current + % ( K )
                    ui:SetText( math.formatK( current ) .. " " .. math.round( percent ) .. "%" )
                end,
                [ 8 ] = function( current, max, percent ) --Current / Max ( K )
                    ui:SetText( math.formatK( current ) .. "/" .. math.formatK( max ) )
                end,
                [ 9 ] = function( current, max, percent ) --Current / Max + % ( K )
                    ui:SetText( math.formatK( current ) .. "/" .. math.formatK( max ) .. " " .. math.round( percent ) .. "%" )
                end
            }

            main.Pos[ Group.Settings.Frame.Morale_Text.anchor ]( ui, Group.Settings.Frame.Morale_Text.x, Group.Settings.Frame.Morale_Text.y, ui:GetSize() )

            function ui:SizeChanged()
                local settings = Group.Settings

                main.Pos[ settings.Frame.Morale_Text.anchor ]( self, settings.Frame.Morale_Text.x, settings.Frame.Morale_Text.y, self:GetSize() )
            end
        end )

        main.name = UI:Create( "TEXT", main, function( ui )
            ui:SetMultiline( false )
            ui:SetSize( Group.Settings.name_width, Group.Font[ Group.Settings.player_font_type ].size[ Group.Settings.player_font_size ] )
            ui:SetVisible( Group.Settings.enable_player_name )
            ui:SetMouseVisible( false )

            if Group.Settings.Frame.Name.anchor == 1 or Group.Settings.Frame.Name.anchor == 4 then
                ui:SetTextAlignment( 4 )
            elseif Group.Settings.Frame.Name.anchor == 5 then
                ui:SetTextAlignment( 5 )
            else
                ui:SetTextAlignment( 6 )
            end

            ui:SetFont( Group:GetFont( Group.Settings.player_font_type, Group.Settings.player_font_size ) )

            if Group.Settings.player_font_outline then
                ui:SetFontStyle( Turbine.UI.FontStyle.Outline )
            else
                ui:SetFontStyle( Turbine.UI.FontStyle.None )
            end
            
            if Group.Settings.enable_unique_player_name_color then
                ui:SetForeColor( Group.Settings.Color.Unique_Name )
                ui:SetOutlineColor( Group.Settings.Color.Unique_Name_Outline )
            else
                ui:SetForeColor( Group.Settings.Color.Name )
                ui:SetOutlineColor( Group.Settings.Color.Name_Outline )
            end

            main.Pos[ Group.Settings.Frame.Name.anchor ]( ui, Group.Settings.Frame.Name.x, Group.Settings.Frame.Name.y, ui:GetSize() )

            function ui:SizeChanged()
                local settings = Group.Settings

                main.Pos[ settings.Frame.Name.anchor ]( self, settings.Frame.Name.x, settings.Frame.Name.y, self:GetSize() )
            end
        end )

        main.class = UI:Create( "BOX", main, function( ui )
            ui:SetSize( 20, 20 )
            ui:SetStretchMode( 1 )
            ui:SetSize( Group.Settings.class_icon_size, Group.Settings.class_icon_size )
            ui:SetColor( 0, 0, 0, 0 )
            ui:SetBlendMode( Turbine.UI.BlendMode.AlphaBlend )
            ui:SetBackColorBlendMode( Turbine.UI.BlendMode.Multiply )
            ui:SetBackground( Group.ClassIcon[ Player:GetClass() ] )
            ui:SetVisible( Group.Settings.enable_class_icon )
            ui:SetMouseVisible( false )

            main.Pos[ Group.Settings.Frame.Class.anchor ]( ui, Group.Settings.Frame.Class.x, Group.Settings.Frame.Class.y, Group.Settings.class_icon_size, Group.Settings.class_icon_size )

            function ui:SizeChanged()
                local settings = Group.Settings

                main.Pos[ settings.Frame.Class.anchor ]( self, settings.Frame.Class.x, settings.Frame.Class.y, settings.class_icon_size, settings.class_icon_size )
            end
        end )

        main.leader = UI:Create( "BOX", main, function( ui )
            ui:SetSize( 16, 16 )
            ui:SetStretchMode( 1 )
            ui:SetSize( Group.Settings.leader_icon_size, Group.Settings.leader_icon_size )
            ui:SetColor( 0, 0, 0, 0 )
            ui:SetBackground( 0x410081C3 )
            ui:SetBlendMode( Turbine.UI.BlendMode.AlphaBlend )
            ui:SetBackColorBlendMode( Turbine.UI.BlendMode.Multiply )
            ui:SetVisible( false )
            ui:SetMouseVisible( false )
            
            main.Pos[ Group.Settings.Frame.Leader.anchor ]( ui, Group.Settings.Frame.Leader.x, Group.Settings.Frame.Leader.y, Group.Settings.leader_icon_size, Group.Settings.leader_icon_size )

            function ui:SizeChanged()
                local settings = Group.Settings

                main.Pos[ settings.Frame.Leader.anchor ]( self, settings.Frame.Leader.x, settings.Frame.Leader.y, settings.leader_icon_size, settings.leader_icon_size )
            end
        end )

        main.special_text = UI:Create( "TEXT", main, function( ui )
            ui:SetSize( 256, Group.Font[ Group.Settings.special_effect_font_type ].size[ Group.Settings.special_effect_font_size ] )
            ui:SetTextAlignment( 6 )
            ui:SetMultiline( false )
            ui:SetMouseVisible( false )
            ui:SetVisible( false )

            ui:SetFont( Group:GetFont( Group.Settings.special_effect_font_type, Group.Settings.special_effect_font_size ) )

            if Group.Settings.special_effect_font_outline then
                ui:SetFontStyle( Turbine.UI.FontStyle.Outline )
            else
                ui:SetFontStyle( Turbine.UI.FontStyle.None )
            end

            ui:SetForeColor( Group.Settings.Color.Special_Effect )
            ui:SetText( "00:00" )
        end )

        main.special = UI:Create( "BOX", main, function( ui, parent )
            ui:SetSize( 32, 32 )
            ui:SetColor( 0, 0, 0, 0 )
            ui:SetBackColorBlendMode( Turbine.UI.BlendMode.AlphaBlend )
            ui:SetMouseVisible( false )
            ui:SetVisible( false )

            function ui:PositionChanged()
                local x, y = self:GetPos()

                local settings = Group.Settings

                if settings.Frame.Special_Effect.anchor == 1 or settings.Frame.Special_Effect.anchor == 4 then
                    parent.special_text:SetTextAlignment( 4 )

                    parent.special_text:SetPos( x + settings.special_effect_size + UI.Spacing, 
                    y + settings.special_effect_size / 2 - Group.Font[ settings.special_effect_font_type ].size[ settings.special_effect_font_size ] / 2 )
                else
                    parent.special_text:SetTextAlignment( 6 )

                    parent.special_text:SetPos( x - 256 - UI.Spacing, y + settings.special_effect_size / 2 - Group.Font[ settings.special_effect_font_type ].size[ settings.special_effect_font_size ] / 2 )
                end
            end

            function ui:SizeChanged()
                local settings = Group.Settings

                main.Pos[ settings.Frame.Special_Effect.anchor ]( self, settings.Frame.Special_Effect.x, settings.Frame.Special_Effect.y, settings.special_effect_size, settings.special_effect_size )
            end

            main.Pos[ Group.Settings.Frame.Special_Effect.anchor ]( ui, Group.Settings.Frame.Special_Effect.x, Group.Settings.Frame.Special_Effect.y, Group.Settings.special_effect_size, Group.Settings.special_effect_size )
        end )

        main.dead_text = UI:Create( "TEXT", main, function( ui )
            ui:SetPos( UI.Spacing, Group.Settings.vitals_y - 15 - UI.Spacing )
            ui:SetSize( Group.Settings.vitals_x, 15 )
            ui:SetFont( Turbine.UI.Lotro.Font.FixedSys15 )
            ui:SetFontStyle( Turbine.UI.FontStyle.Outline )
            ui:SetText( "Defeated" )
            ui:SetVisible( false )
            ui:SetMouseVisible( false )
        end )

        main.link_dead = UI:Create( "TEXT", main, function( ui )
            ui:SetPos( UI.Spacing, Group.Settings.vitals_y - 15 - UI.Spacing )
            ui:SetSize( Group.Settings.vitals_x, 15 )
            ui:SetFont( Turbine.UI.Lotro.Font.FixedSys15 )
            ui:SetFontStyle( Turbine.UI.FontStyle.Outline )
            ui:SetText( "Offline" )
            ui:SetMouseVisible( false )
            ui:SetVisible( false )
        end )

        main.effect = {}

        for i = 1, main.effect_count do
            main.effect[ i ] = UI:Create( "BOX", main, function( ui )
                ui:SetColor( 255, 255, 255, 0 )
                ui:SetBackColorBlendMode( Turbine.UI.BlendMode.AlphaBlend )
                ui:SetMouseVisible( false )
                ui:SetVisible( false )

                ui.start    = main.effect_count - i
                ui.duration = 0

                --[[ local tbl = {
                    [ Turbine.Gameplay.EffectCategory.Disease ]    = 0x410F25C3, --Disease
                    [ Turbine.Gameplay.EffectCategory.Wound ]      = 0x410F25C4, --Wound
                    [ Turbine.Gameplay.EffectCategory.Fear ]       = 0x410F25C5, --Fear
                    [ Turbine.Gameplay.EffectCategory.Poison ]     = 0x410F25C6, --Poison
                    [ Turbine.Gameplay.EffectCategory.Corruption ] = 0x410F25C2 --Corruption
                } ]]

                local tbl = {
                    [ Turbine.Gameplay.EffectCategory.Disease ]    = Color( 255, 255, 0 ), --Disease
                    [ Turbine.Gameplay.EffectCategory.Wound ]      = Color( 255, 0, 0 ), --Wound
                    [ Turbine.Gameplay.EffectCategory.Fear ]       = Color( 255, 0, 255 ), --Fear
                    [ Turbine.Gameplay.EffectCategory.Poison ]     = Color( 0, 255, 0 ), --Poison
                    [ Turbine.Gameplay.EffectCategory.Corruption ] = Color( 0, 0, 255 ) --Corruption
                }

                function ui:Activate( effect )
                    if effect:IsCurable() then
                        self.border:SetVisible( true )

                        --self.border:SetBackground( tbl[ effect:GetCategory() ] )

                        self.border:SetBorderColor( tbl[ effect:GetCategory() ]  )
                    else
                        self.border:SetVisible( false )
                    end

                    local settings = Group.Settings
                    
                    self:SetBackground( effect:GetIcon() )
                    self:SetStretchMode( 1 )
                    self:SetSize( settings.effect_size, settings.effect_size )

                    self:SetVisible( true )

                    if not self.static then
                        self.text:SetText( self.duration )
                        self.text:SetVisible( settings.enable_effect_duration )
                    else
                        self.text:SetVisible( false )
                    end
                end

                ui.text = UI:Create( "TEXT", ui, function( ui )
                    ui:SetSize( 32, 32 )
                    ui:SetTextAlignment( 5 )
                    ui:SetMultiline( false )
                    ui:SetVisible( Group.Settings.enable_effect_duration )
                    ui:SetFont( Group:GetFont( Group.Settings.effect_font_type, Group.Settings.effect_font_size ) )
                    ui:SetMouseVisible( false )

                    ui:SetText( i )

                    if Group.Settings.effect_font_outline then
                        ui:SetFontStyle( Turbine.UI.FontStyle.Outline )
                    else
                        ui:SetFontStyle( Turbine.UI.FontStyle.None )
                    end
                end )

                ui.border = UI:Create( "BORDER", ui, function( ui )
                    --[[ ui:SetColor( 0, 0, 0, 0 )
                    --ui:SetBackground( "Darf/MiniRaid/Resources/corruption.tga" )
                    ui:SetBlendMode( Turbine.UI.BlendMode.AlphaBlend )
                    ui:SetBackColorBlendMode( Turbine.UI.BlendMode.Multiply )
                    ui:SetSize( 32, 32 ) ]]

                    ui:SetSize( 32, 32 )
                    ui:SetBorderSize( 2 )
                    --ui:SetBorderColor( tbl[ Turbine.Gameplay.EffectCategory.Corruption ] )
                end )

                ui:SetSize( 32, 32 )
                ui:SetStretchMode( 1 )
                ui:SetSize( Group.Settings.effect_size, Group.Settings.effect_size )
            end )
        end

        main.selection = UI:Create( "BORDER", main, function( ui )
            ui:SetSize( Group.Settings.vitals_x, Group.Settings.vitals_y )
            ui:SetBorderSize( Group.Settings.selection_size )
            ui:SetBorderColor( Group.Settings.Color.Select )
            ui:SetVisible( false )
        end )

        main.entity = UI:Create( "ENTITY", main, function( ui )
            ui:SetSize( Group.Settings.vitals_x, Group.Settings.vitals_y )
        end )

        function main:PositionChanged()
            local settings = Group.Settings

            self.class:SetStretchMode( 1 )
            self.class:SetSize( settings.class_icon_size, settings.class_icon_size )

            self.leader:SetStretchMode( 1 )
            self.leader:SetSize( settings.leader_icon_size, settings.leader_icon_size )

            for i = 1, self.effect_count do
                self.effect[ i ]:SetStretchMode( 1 )
                self.effect[ i ]:SetSize( settings.effect_size, settings.effect_size )
            end
        end

        function main:SizeChanged()
            local w, h = self:GetSize()

            local settings = Group.Settings

            if settings.enable_classic_power_bar then
                self.morale:SetPos( 0, 0 )
                self.morale:SetHeight( h - settings.power_size )

                self.bubble:SetHeight( h - settings.power_size )

                self.power:SetHeight( settings.power_size )
                self.power:SetPos( 0, h - settings.power_size )
            else
                self.morale:SetPos( settings.power_size, 0 )
                self.morale:SetHeight( h )
            
                self.bubble:SetHeight( h )

                self.power:SetWidth( settings.power_size )
            end

            self.selection:SetSize( w, h )

            self.Pos[ settings.Frame.Name.anchor ]( self.name, settings.Frame.Name.x, settings.Frame.Name.y, self.name:GetSize() )
            self.Pos[ settings.Frame.Class.anchor ]( self.class, settings.Frame.Class.x, settings.Frame.Class.y, settings.class_icon_size, settings.class_icon_size )
            self.Pos[ settings.Frame.Leader.anchor ]( self.leader, settings.Frame.Leader.x, settings.Frame.Leader.y, self.leader:GetSize() )
            self.Pos[ settings.Frame.Morale_Text.anchor ]( self.morale.text, settings.Frame.Morale_Text.x, settings.Frame.Morale_Text.y, self.morale.text:GetSize() )
            self.Pos[ settings.Frame.Special_Effect.anchor ]( self.special, settings.Frame.Special_Effect.x, settings.Frame.Special_Effect.y, settings.special_effect_size, settings.special_effect_size )

            self.power.text.Pos[ settings.Frame.Power_Text.anchor ]( self.power.text, settings.Frame.Power_Text.x, settings.Frame.Power_Text.y, self.power.text:GetSize() )

            self.dead:SetSize( w, h )

            self.dead_text:SetPos( UI.Spacing, h - 15 - UI.Spacing )
            self.dead_text:SetSize( w, 15 )

            self.link_dead:SetPos( UI.Spacing, h - 15 - UI.Spacing )
            self.link_dead:SetSize( w, 15 )

            self.entity:SetSize( w, h )

            self:PositionChanged()

            self:Init()

            self:EffectLayout()
        end

        function main:EffectLayout()
            local count    = 1
            local settings = Group.Settings

            local row_x = settings.effect_row_count
            local row_y = math.ceil( self.effect_count / settings.effect_row_count )

            local w
            local h
            
            if settings.enable_classic_power_bar then
                w = settings.vitals_x
                h = settings.vitals_y - settings.power_size
            else
                w = settings.vitals_x - settings.power_size
                h = settings.vitals_y
            end

            table.sort( self.effect, function( a, b ) return ( a.start + a.duration ) > ( b.start + b.duration ) end )

            for y = 1, row_y do
                for x = 1, row_x do
                    local offset_x, offset_y = self.EffectPos[ settings.Frame.Effect.anchor ]( x, y, w, h, settings.Frame.Effect.x, settings.Frame.Effect.y )

                    self.effect[ count ]:SetPos( offset_x, offset_y )
                    
                    count = count + 1
    
                    if count > self.effect_count then return end
                end
            end
        end

        function main:AddEffect( effect, init, disablelayout )
            local id       = effect:GetStartTime()
            local name     = effect:GetName()
            local duration = effect:GetDuration()

            self.effect_list[ name ] = self.effect_list[ name ] or {}

            if not init then
                for i = 1, #self.effect_list[ name ] do
                    if self.effect_list[ name ][ i ] == id then return end
                end
            end

            local count = #self.effect_list[ name ]

            self.effect_list[ name ][ count + 1 ] = id

            count = count + 1

            if count > Group.Settings.effect_stack[ name ] then
                for i = 1, #self.effect do
                    if self.effect[ i ].active and self.effect[ i ].start == self.effect_list[ name ][ 1 ] and self.effect[ i ].name == name then
                        self.effect[ i ]:SetVisible( false )
    
                        self.effect[ i ].active = false
    
                        break
                    end
                end

                table.remove( self.effect_list[ name ], 1 )

                for i = 1, self.effect_count do
                    if not self.effect[ i ].active then
                        self.effect[ i ].name     = name
                        self.effect[ i ].start    = id
                        self.effect[ i ].duration = duration
    
                        if duration > 86400 then
                            self.effect[ i ].static = true
                        else
                            self.effect[ i ].static = false
                        end
    
                        self.effect[ i ].active = true
    
                        self.effect[ i ]:Activate( effect )
                        
                        if not disablelayout then
                            self:EffectLayout()
                        end
    
                        break
                    end
                end
            else
                for i = 1, self.effect_count do
                    if not self.effect[ i ].active then
                        self.effect[ i ].name     = name
                        self.effect[ i ].start    = id
                        self.effect[ i ].duration = duration
    
                        if duration > 86400 then
                            self.effect[ i ].static = true
                        else
                            self.effect[ i ].static = false
                        end
    
                        self.effect[ i ].active = true
    
                        self.effect[ i ]:Activate( effect )
    
                        if not disablelayout then
                            self:EffectLayout()
                        end
    
                        break
                    end
                end
            end
        end

        function main:RefreshEffects( list, clear )
            if clear then
                for i = 1, self.effect_count do
                    self.effect[ i ]:SetVisible( false )

                    self.effect[ i ].start    = 0
                    self.effect[ i ].duration = 0

                    self.effect[ i ].active = false
                end
            end

            local settings = Group.Settings
            
            if not settings.enable_effect then return end

            for i = 1, list:GetCount() do
                local effect = list:Get( i )
                local name   = effect:GetName()

                if settings.special_effect == name then
                    self:AddSpecialEffect( effect )
                end

                if self.localplayer then
                    if settings.track_all then
                        if not settings.effect_blacklist[ name ] then
                            self:AddLocalEffect( effect, true )
                        end
                    else
                        if settings.effect_self[ name ] then
                            self:AddLocalEffect( effect, true )
                        else
                            if settings.track_curable and effect:IsCurable() then
                                if not settings.effect_blacklist[ name ] then
                                    self:AddLocalEffect( effect, true )
                                end
                            elseif settings.track_debuff and effect:IsDebuff() then
                                if not settings.effect_blacklist[ name ] then
                                    self:AddLocalEffect( effect, true )
                                end
                            elseif settings.track_crowd_control and Group.CC[ name ] then
                                if not settings.effect_blacklist[ name ] then
                                    self:AddLocalEffect( effect, true )
                                end
                            end
                        end
                    end
                else
                    if settings.effect[ name ] then
                        self:AddEffect( effect, true, true )
                    end
                end
            end

            self:EffectLayout()
        end

        function main:AddLocalEffect( effect, disablelayout )
            for i = 1, self.effect_count do
                if not self.effect[ i ].active then
                    self.effect[ i ].name     = effect:GetName()
                    self.effect[ i ].start    = effect:GetStartTime()

                    local duration = effect:GetDuration()
                    
                    self.effect[ i ].duration = duration

                    if duration > 86400 then
                        self.effect[ i ].static = true
                    else
                        self.effect[ i ].static = false
                    end

                    self.effect[ i ].active = true

                    self.effect[ i ]:Activate( effect )

                    if not disablelayout then
                        self:EffectLayout()
                    end

                    break
                end
            end
        end

        function main:RemoveLocalEffect( effect )
            local name = effect:GetName()

            for i = 1, #self.effect do
                if self.effect[ i ].active and self.effect[ i ].start == effect:GetStartTime() and self.effect[ i ].name == name then
                    self.effect[ i ]:SetVisible( false )

                    self.effect[ i ].start    = 0
                    self.effect[ i ].duration = 0

                    self.effect[ i ].active = false

                    self:EffectLayout()

                    break
                end
            end
        end

        function main:AddSpecialEffect( effect )
            self.special.start    = effect:GetStartTime()
            self.special.duration = effect:GetDuration()

            local settings = Group.Settings

            self.special:SetBackground( effect:GetIcon() )
            self.special:SetStretchMode( 1 )
            self.special:SetSize( settings.special_effect_size, settings.special_effect_size )

            self.special_text:SetText( self.special.duration )

            if settings.enable_special_effect then
                self.special:SetVisible( true )
                self.special_text:SetVisible( settings.enable_special_effect_duration )
            end

            self.special.active = true
        end

        function main:VisibleChanged()
            self:SetWantsUpdates( self:IsVisible() )
        end

        function main:Update()
            local settings = Group.Settings

            if self:IsHovering() then
                self.morale:SetColor( settings.Color.Hover )
            else
                local percent = self.player_morale / self.player_morale_max

                if percent > settings.morale_threshold then
                    if settings.enable_unique_player_morale_color and self.localplayer then
                        self.morale:SetColor( settings.Color.Unique_Morale )
                    else
                        if settings.enable_unique_class_morale_color then
                            self.morale:SetColor( settings.Color.Class[ self.player_class ] )
                        else
                            self.morale:SetColor( settings.Color.Morale )
                        end
                    end
                else
                    self.morale:SetColor( settings.Color.Threshold )
                end
            end

            local currentTime = Turbine.Engine.GetGameTime()

            if self.special.active then
                local endTime = self.special.start + self.special.duration

                local time = math.round( self.special.duration ) - math.ceil( currentTime - self.special.start )

                if currentTime > endTime then
                    self.special:SetVisible( false )

                    self.special_text:SetVisible( false )

                    self.special.active = false
                else
                    self.special_text:SetText( " " .. string.formatSeconds( time ) )
                end
            end

            local speed = 3
            local range = 0.3
            local start = 0.5

            for i = 1, self.effect_count do
                if self.effect[ i ].active then
                    local endTime = self.effect[ i ].start + self.effect[ i ].duration

                    if not self.effect[ i ].static then
                        local time = math.round( self.effect[ i ].duration ) - math.ceil( currentTime - self.effect[ i ].start )

                        if time < 0 then time = 0 end

                        if time > settings.effect_duration_threshold then 
                            self.effect[ i ].text:SetText( "" ) 
                        else
                            self.effect[ i ].text:SetText( time )
                        end

                        if time <= settings.effect_flash_threshold then 
                            local pulse = math.sin( currentTime * speed ) * range + start

                            self.effect[ i ]:SetOpacity( pulse )
                        else
                            self.effect[ i ]:SetOpacity( 1 )
                        end
                    end
    
                    if currentTime > endTime then
                        if not self.localplayer then
                            for k = 1, #self.effect_list[ self.effect[ i ].name ] do
                                if self.effect_list[ self.effect[ i ].name ][ k ] == self.effect[ i ].start then
                                    table.remove( self.effect_list[ self.effect[ i ].name ], k )

                                    break
                                end
                            end
                        end

                        self.effect[ i ]:SetVisible( false )

                        self.effect[ i ].start    = 0
                        self.effect[ i ].duration = 0

                        self.effect[ i ].active = false

                        self:EffectLayout()
                    end
                end
            end
        end

        function main:SetSelected( bool )
            self.selection:SetVisible( bool )

            return self
        end

        function main:SetDead( dead )
            local settings = Group.Settings

            self.dead:SetVisible( dead )
            self.dead_text:SetVisible( dead )

            if dead then
                self.power.text:SetVisible( false )
                self.morale.text:SetVisible( false )

                if not ( settings.special_effect == "Do Not Fall This Day" ) then
                    self.special:SetVisible( false )
                end

                self.special.active = false

                self.special_text:SetVisible( false )

                for i = 1, self.effect_count do
                    if not ( self.effect[ i ].name == "Do Not Fall This Day" ) then
                        self.effect[ i ]:SetVisible( false )
                    end

                    self.effect[ i ].start    = 0
                    self.effect[ i ].duration = 0

                    self.effect[ i ].active = false
                end
            else
                self.power.text:SetVisible( settings.enable_power_text )
                self.morale.text:SetVisible( settings.enable_morale_text )

                if settings.special_effect == "Do Not Fall This Day" then
                    self.special:SetVisible( false )
                end

                for i = 1, self.effect_count do
                    if self.effect[ i ].name == "Do Not Fall This Day" then
                        self.effect[ i ]:SetVisible( false )

                        break
                    end
                end
            end
        end

        function main:SetName( str )
            self.name:SetText( str )
        end

        function main:GetName()
            return self.player_name
        end

        function main:SetClass( class )
            self.class:SetBackground( Group.ClassIcon[ class ] )

            if class == Turbine.Gameplay.Class.Beorning then
                self.power:SetColor( Group.Settings.Color.Wrath )
            end
        end

        function main:SetLeader( bool )
            self.leader:SetVisible( bool )
        end

        function main:SetOffline( bool )
            self.link_dead:SetVisible( bool )

            self.morale:SetVisible( not bool )
            self.power:SetVisible( not bool )

            local settings = Group.Settings

            if bool then
                self.dead:SetVisible( false )
                self.dead_text:SetVisible( false )

                self.morale:SetWidth( 0 )

                if settings.enable_classic_power_bar then
                    self.power:SetWidth( 0 )
                else
                    self.power:SetHeight( 0 )
                end

                self.special:SetVisible( false )
                self.special_text:SetVisible( false )

                self.power.text:SetVisible( false )
                self.morale.text:SetVisible( false )

                for i = 1, self.effect_count do
                    self.effect[ i ]:SetVisible( false )

                    self.effect[ i ].start    = 0
                    self.effect[ i ].duration = 0
                    
                    self.effect[ i ].active = false
                end
            else
                self.power.text:SetVisible( settings.enable_power_text )
                self.morale.text:SetVisible( settings.enable_morale_text )

                self:Init()
            end
        end

        function main:GetPlayer()
            return self.player
        end

        function main:SetEntity( entity )
            self.entity:SetEntity( entity )
        end

        function main:RemoveCallbacks()
            if self.player then
                for k, v in pairs( self.callback ) do
                    RemoveCallback( self.player, k, v )
                end
        
                self.callback = {}
            end

            if self.player_effects then
                for k, v in pairs( self.callback_effect ) do
                    RemoveCallback( self.player_effects, k, v )
                end
        
                self.callback_effect = {}
            end
        end

        function main:SetPlayer( player )
            if not player then return end
            
            self:RemoveCallbacks()
            
            self.player         = player
            self.player_name    = self.player:GetName()
            self.player_class   = self.player:GetClass()
            self.player_effects = self.player:GetEffects()

            self.isDead = false

            if self.player_name == Player:GetName() then
                self.localplayer = true
            end

            self:SetName( self.player_name )
            self:SetEntity( self.player )
            self:SetClass( self.player_class )

            local settings = Group.Settings

            if self.localplayer and settings.enable_unique_player_name_color then
                self.name:SetForeColor( settings.Color.Unique_Name )
            else
                self.name:SetForeColor( settings.Color.Name )
            end

            self.callback[ "MoraleChanged" ]          = AddCallback( self.player, "MoraleChanged", function() self:MoraleChanged() end )
            self.callback[ "MaxMoraleChanged" ]       = AddCallback( self.player, "MaxMoraleChanged", function() self:MaxMoraleChanged() end )
            self.callback[ "PowerChanged" ]           = AddCallback( self.player, "PowerChanged", function() self:PowerChanged() end )
            self.callback[ "MaxPowerChanged" ]        = AddCallback( self.player, "MaxPowerChanged", function() self:MaxPowerChanged() end )
            self.callback[ "TemporaryMoraleChanged" ] = AddCallback( self.player, "TemporaryMoraleChanged", function() self:BubbleChanged() end )

            self.callback[ "IsLinkDeadChanged" ]      = AddCallback( self.player, "IsLinkDeadChanged", function() self:SetOffline( self.player:IsLinkDead() ) end )

            self:RefreshEffects( self.player_effects )
            
            if self.localplayer then
                self.callback_effect[ "EffectAdded" ] = AddCallback( self.player_effects, "EffectAdded", function( effectList, effect )
                    local effect   = effectList:Get( effect.Index )
                    local name     = effect:GetName()
                    local settings = Group.Settings
    
                    if settings.special_effect == name then
                        self:AddSpecialEffect( effect )
                    end

                    local group = Group

                    if group.record then
                        if group.record_list[ name ] then
                            group.record_list[ name ] = group.record_list[ name ] + 1
                        else
                            group.record_list[ name ] = 1
                        end
                    end

                    if not settings.enable_effect then return end

                    if settings.track_all then
                        if not settings.effect_blacklist[ name ] then
                            self:AddLocalEffect( effect )
                        end
                    else
                        if settings.effect_self[ name ] then
                            self:AddLocalEffect( effect )
                        else
                            if settings.track_curable and effect:IsCurable() then
                                if not settings.effect_blacklist[ name ] then
                                    self:AddLocalEffect( effect )
                                end
                            elseif settings.track_debuff and effect:IsDebuff() then
                                if not settings.effect_blacklist[ name ] then
                                    self:AddLocalEffect( effect )
                                end
                            elseif settings.track_crowd_control and Group.CC[ name ] then
                                if not settings.effect_blacklist[ name ] then
                                    self:AddLocalEffect( effect )
                                end
                            end
                        end
                    end
                end )

                self.callback_effect[ "EffectRemoved" ] = AddCallback( self.player_effects, "EffectRemoved", function( effectList, effect )
                    local effect   = effect.Effect
                    local name     = effect:GetName()
                    local settings = Group.Settings

                    if not settings.enable_effect then return end
    
                    if settings.track_all then
                        if not settings.effect_blacklist[ name ] then
                            self:RemoveLocalEffect( effect )
                        end
                    else
                        if settings.effect_self[ name ] then
                            self:RemoveLocalEffect( effect )
                        else
                            if settings.track_curable and effect:IsCurable() then
                                if not settings.effect_blacklist[ name ] then
                                    self:RemoveLocalEffect( effect )
                                end
                            elseif settings.track_debuff and effect:IsDebuff() then
                                if not settings.effect_blacklist[ name ] then
                                    self:RemoveLocalEffect( effect )
                                end
                            elseif settings.track_crowd_control and Group.CC[ name ] then
                                if not settings.effect_blacklist[ name ] then
                                    self:RemoveLocalEffect( effect )
                                end
                            end
                        end
                    end
                end )
            else
                self.callback_effect[ "EffectAdded" ] = AddCallback( self.player_effects, "EffectAdded", function( effectList, effect )
                    local effect = effectList:Get( effect.Index )
                    local name   = effect:GetName()

                    local settings = Group.Settings
    
                    if settings.special_effect == name then
                        self:AddSpecialEffect( effect )
                    end

                    local group = Group

                    if group.record then
                        if group.record_list[ name ] then
                            group.record_list[ name ] = group.record_list[ name ] + 1
                        else
                            group.record_list[ name ] = 1
                        end
                    end

                    if not settings.enable_effect then return end
    
                    if settings.effect[ name ] then
                        self:AddEffect( effect )
                    end
                end )
            end
        end

        function main:Init()
            self.player_morale     = self.player:GetMorale()
            self.player_morale_max = self.player:GetMaxMorale()
            self.player_power      = self.player:GetPower()
            self.player_power_max  = self.player:GetMaxPower()
            self.player_bubble     = self.player:GetTemporaryMorale()

            local bubble_percent = self.player_bubble / ( self.player_morale_max + self.player_bubble )
            local morale_percent = self.player_morale / self.player_morale_max
            local power_percent  = self.player_power  / self.player_power_max

            local settings = Group.Settings

            local bubble_size = math.round( ( settings.vitals_x - settings.power_size ) * bubble_percent )
            local morale_size = math.round( ( settings.vitals_x - settings.power_size - bubble_size ) * morale_percent )
            local power_size  = math.round( settings.vitals_y * power_percent )

            if settings.enable_morale_text then
                self.morale.text.format[ settings.morale_text_format ]( self.player_morale, self.player_morale_max, morale_percent * 100 )
            end

            if settings.enable_power_text then
                self.power.text.format[ settings.power_text_format ]( self.player_power, self.player_power_max, power_percent * 100 )
            end

            if settings.enable_classic_power_bar then
                bubble_size = math.round( settings.vitals_x * bubble_percent )
                morale_size = math.round( ( settings.vitals_x - bubble_size ) * morale_percent )
                power_size  = math.round( settings.vitals_x * power_percent )

                self.bubble:SetPos( settings.vitals_x - bubble_size, 0 )
                self.bubble:SetWidth( bubble_size )
            else
                self.bubble:SetPos( settings.power_size + settings.vitals_x - settings.power_size - bubble_size, 0 )
                self.bubble:SetWidth( bubble_size )
            end

            self.morale:SetWidth( morale_size )

            if self.player_morale == 0 then
                if settings.enable_classic_power_bar then
                    self.power:SetWidth( 0 )
                else
                    self.power:SetHeight( 0 )
                end

                if not self.isDead then
                    self:SetDead( true )
                end

                self.isDead = true
            else
                if settings.enable_classic_power_bar then
                    self.power:SetWidth( power_size )
                else
                    self.power:SetPos( 0, settings.vitals_y - power_size )
                    self.power:SetHeight( power_size )
                end

                if self.isDead then
                    self:SetDead( false )
                end

                self.isDead = false
            end

            if self.player:IsLinkDead() then
                self:SetOffline( true )
            end
        end

        function main:MoraleChanged()
            self.player_morale = self.player:GetMorale()

            local settings = Group.Settings

            if self.player_morale == 0 then
                if settings.enable_classic_power_bar then
                    self.power:SetWidth( 0 )
                else
                    self.power:SetHeight( 0 )
                end

                if not self.isDead then
                    self:SetDead( true )
                end

                self.isDead = true
            else
                if self.isDead then
                    self:SetDead( false )
                end

                self.isDead = false
            end
    
            local bubble_percent = self.player_bubble / ( self.player_morale_max + self.player_bubble )
            local morale_percent = self.player_morale / self.player_morale_max
    
            local bubble_size = math.round( ( settings.vitals_x - settings.power_size ) * bubble_percent )
            local morale_size = math.round( ( settings.vitals_x - settings.power_size - bubble_size ) * morale_percent )

            if settings.enable_morale_text then
                self.morale.text.format[ settings.morale_text_format ]( self.player_morale, self.player_morale_max, morale_percent * 100 )
            end

            if settings.enable_classic_power_bar then
                bubble_size = math.round( settings.vitals_x * bubble_percent )
                morale_size = math.round( ( settings.vitals_x - bubble_size ) * morale_percent )

                self.bubble:SetPos( settings.vitals_x - bubble_size, 0 )
                self.bubble:SetWidth( bubble_size )
            else
                self.bubble:SetPos( settings.power_size + settings.vitals_x - settings.power_size - bubble_size, 0 )
                self.bubble:SetWidth( bubble_size )
            end
    
            self.morale:SetWidth( morale_size )
        end
        
        function main:MaxMoraleChanged()
            self.player_morale_max = self.player:GetMaxMorale()
    
            local bubble_percent = self.player_bubble / ( self.player_morale_max + self.player_bubble )
            local morale_percent = self.player_morale / self.player_morale_max

            local settings = Group.Settings
    
            local bubble_size = math.round( ( settings.vitals_x - settings.power_size ) * bubble_percent )
            local morale_size = math.round( ( settings.vitals_x - settings.power_size - bubble_size ) * morale_percent )

            if settings.enable_morale_text then
                self.morale.text.format[ settings.morale_text_format ]( self.player_morale, self.player_morale_max, morale_percent * 100 )
            end

            if settings.enable_classic_power_bar then
                bubble_size = math.round( settings.vitals_x * bubble_percent )
                morale_size = math.round( ( settings.vitals_x - bubble_size ) * morale_percent )

                self.bubble:SetPos( settings.vitals_x - bubble_size, 0 )
                self.bubble:SetWidth( bubble_size )
            else
                self.bubble:SetPos( settings.power_size + settings.vitals_x - settings.power_size - bubble_size, 0 )
                self.bubble:SetWidth( bubble_size )
            end
    
            self.morale:SetWidth( morale_size )
        end
        
        function main:PowerChanged()
            if self.isDead then return end

            self.player_power = self.player:GetPower()
    
            local power_percent = self.player_power / self.player_power_max

            local settings = Group.Settings

            if settings.enable_power_text then
                self.power.text.format[ settings.power_text_format ]( self.player_power, self.player_power_max, power_percent * 100 )
            end
    
            if settings.enable_classic_power_bar then
                local power_size = math.round( settings.vitals_x * power_percent )

                self.power:SetWidth( power_size )
            else
                local power_size = math.round( settings.vitals_y * power_percent )

                self.power:SetPos( 0, settings.vitals_y - power_size )
                self.power:SetHeight( power_size )
            end
        end
        
        function main:MaxPowerChanged()
            if self.isDead then return end

            self.player_power_max = self.player:GetMaxPower()
    
            local power_percent = self.player_power / self.player_power_max

            local settings = Group.Settings

            if settings.enable_power_text then
                self.power.text.format[ settings.power_text_format ]( self.player_power, self.player_power_max, power_percent * 100 )
            end
    
            if settings.enable_classic_power_bar then
                local power_size = math.round( settings.vitals_x * power_percent )

                self.power:SetWidth( power_size )
            else
                local power_size = math.round( settings.vitals_y * power_percent )

                self.power:SetPos( 0, settings.vitals_y - power_size )
                self.power:SetHeight( power_size )
            end
        end
    
        function main:BubbleChanged()
            self.player_bubble = self.player:GetTemporaryMorale()
    
            local bubble_percent = self.player_bubble / ( self.player_morale_max + self.player_bubble )
            local morale_percent = self.player_morale / self.player_morale_max

            local settings = Group.Settings
    
            local bubble_size = math.round( ( settings.vitals_x - settings.power_size ) * bubble_percent )
            local morale_size = math.round( ( settings.vitals_x - settings.power_size - bubble_size ) * morale_percent )

            if settings.enable_classic_power_bar then
                bubble_size = math.round( settings.vitals_x * bubble_percent )
                morale_size = math.round( ( settings.vitals_x - bubble_size ) * morale_percent )

                self.bubble:SetPos( settings.vitals_x - bubble_size, 0 )
                self.bubble:SetWidth( bubble_size )
            else
                self.bubble:SetPos( settings.power_size + settings.vitals_x - settings.power_size - bubble_size, 0 )
                self.bubble:SetWidth( bubble_size )
            end
    
            self.morale:SetWidth( morale_size )
        end
    end
end )

function Group:DefineGridLayout( count )
    local settings = Group.Settings

    if count <= 3 then
        self.Grid_Size_X = settings.Layout[ 1 ].x
        self.Grid_Size_Y = settings.Layout[ 1 ].y
    elseif count <= 6 then
        self.Grid_Size_X = settings.Layout[ 2 ].x
        self.Grid_Size_Y = settings.Layout[ 2 ].y
    elseif count <= 12 then
        self.Grid_Size_X = settings.Layout[ 3 ].x
        self.Grid_Size_Y = settings.Layout[ 3 ].y
    elseif count <= 18 then
        self.Grid_Size_X = settings.Layout[ 4 ].x
        self.Grid_Size_Y = settings.Layout[ 4 ].y
    elseif count <= 24 then
        self.Grid_Size_X = settings.Layout[ 5 ].x
        self.Grid_Size_Y = settings.Layout[ 5 ].y
    end
end

function Group:RefreshLayout()
    local slot = 1

    local w, h = PartyVitals:GetSize()

    for y = 1, self.Grid_Size_Y do
        for x = 1, self.Grid_Size_X do
            local count    = self.PartyCount - self.Grid_Size_X * ( y - 1 )
            local row_size = self.Grid_Size_X

            if y == self.Grid_Size_Y and self.Grid_Size_X < count then
                row_size = count - self.Grid_Size_X
            elseif ( y == self.Grid_Size_Y and self.Grid_Size_X > count ) or self.Grid_Size_X > count then
                row_size = count
            end

            local pos_x
            local pos_y

            if self.Settings.horizontal_layout_direction == 1 then --Left
                pos_x = w / 2 + self.Settings.vitals_x * ( x - 1 ) + self.Settings.vitals_external_spacing * ( x - 1 )
            elseif self.Settings.horizontal_layout_direction == 2 then --Center
                pos_x = w / 2 + self.Settings.vitals_x * ( x - 1 ) + self.Settings.vitals_external_spacing * ( x - 1 ) - ( self.Settings.vitals_x * row_size + self.Settings.vitals_external_spacing * ( row_size - 1 ) ) / 2
            else --Right
                pos_x = w / 2 - self.Settings.vitals_x * x - self.Settings.vitals_external_spacing * ( x - 1 )
            end

            if self.Settings.vertical_layout_direction == 1 then --Top
                pos_y = h / 2 + self.Settings.vitals_y * ( y - 1 ) + self.Settings.vitals_external_spacing * ( y - 1 )
            else --Bottom
                pos_y = h / 2 - self.Settings.vitals_y * y - self.Settings.vitals_external_spacing * ( y - 1 )
            end

            self.Party[ slot ]:SetPos( pos_x, pos_y )

            slot = slot + 1

            if slot > self.PartyCount then return end
        end
    end
end

function Group:Create( single )
    if single then
        self.PartyCount = 1

        local member = Player
        local name   = member:GetName()

        self.Party[ 1 ] = UI:Create( "FRAME", PartyVitals, function( ui )
            PartyVitals:AddDraggable( ui )

            ui:SetPlayer( member )
        end )

        self.PartyName[ name ] = 1

        self.Leader = name

        if self.Settings.enable_leader_icon then
            self.Party[ 1 ]:SetLeader( true )
        end

        local target = Player:GetTarget()

        if target then
            if target:GetName() == name then
                self.Selected = self.Party[ 1 ]:SetSelected( true )
            end
        end

        self.Party[ 1 ]:Init()
    else
        local party = Player:GetParty()

        self.PartyCount = party:GetMemberCount()

        function party:LeaderChanged()
            local name = self:GetLeader():GetName()
            
            if Group.Settings.enable_leader_icon then
                Group.Party[ Group.PartyName[ name ] ]:SetLeader( true )
                Group.Party[ Group.PartyName[ Group.Leader ] ]:SetLeader( false )
            end

            Group.Leader = name
        end

        local tbl   = {}
        local name  = {}
        local check = {}

        for i = 1, self.PartyCount do
            local member = party:GetMember( i )
            local str    = member:GetName()

            name[ str ] = member
        end

        if self.Layout then
            for i = 1, #self.Layout do
                if name[ self.Layout[ i ] ] then
                    check[ self.Layout[ i ] ] = true

                    table.insert( tbl, name[ self.Layout[ i ] ] )
                end
            end

            Turbine.PluginData.Save( Turbine.DataScope.Character, "MiniRaid_Layout", false )
        end

        for i = 1, self.PartyCount do
            local member = party:GetMember( i )

            if not check[ member:GetName() ] then
                table.insert( tbl, member )
            end
        end

        for i = 1, self.PartyCount do
            local member = tbl[ i ]
            local name   = member:GetName()
    
            self.Party[ i ] = UI:Create( "FRAME", PartyVitals, function( ui )
                PartyVitals:AddDraggable( ui )
    
                ui:SetPlayer( member )
            end )
    
            self.PartyName[ name ] = i
    
            if member == party:GetLeader() then 
                self.Leader = name
    
                if self.Settings.enable_leader_icon then
                    self.Party[ i ]:SetLeader( true )
                end
            end
    
            local target = Player:GetTarget()
    
            if target then
                if target:GetName() == name then
                    self.Selected = self.Party[ i ]:SetSelected( true )
                end
            end
    
            self.Party[ i ]:Init()
        end
    end

    self:DefineGridLayout( self.PartyCount )
    self:RefreshLayout()
end

function Group:Remove()
    self.PartyName  = {}
    self.PartyCount = 0

    for i = 1, #self.Party do
        self.Party[ i ]:RemoveCallbacks()
        self.Party[ i ]:SetWantsUpdates( false )
        self.Party[ i ]:SetVisible( false )
    end

    self.Party = {}

    Turbine.PluginData.Save( Turbine.DataScope.Character, "MiniRaid_Layout", false )
end

function Group:RemoveMember( name )
    local index = self.PartyName[ name ]

    self.Party[ index ]:RemoveCallbacks()
    self.Party[ index ]:SetWantsUpdates( false )
    self.Party[ index ]:SetVisible( false )

    table.remove( self.Party, index )

    self.PartyName = {}

    for i = 1, #self.Party do
        self.PartyName[ self.Party[ i ].player_name ] = i
    end

    self.PartyCount = self.PartyCount - 1

    self:DefineGridLayout( self.PartyCount )
    self:RefreshLayout()
end