UI.BASE   = {}
UI.ACTIVE = {}
UI.OBJECT = {
    BASE      = function() return Turbine.UI.Window() end,
    TEXT      = function() return Turbine.UI.Label() end,
    LIST      = function() return Turbine.UI.ListBox() end,
    SCROLL    = function() return Turbine.UI.ScrollBar() end,
    BUTTON    = function() return Turbine.UI.Button() end,
    ENTITY    = function() return Turbine.UI.Lotro.EntityControl() end,
    EFFECT    = function() return Turbine.UI.Lotro.EffectDisplay() end,
    WINDOW    = function() return Turbine.UI.Lotro.Window() end,
    LBUTTON   = function() return Turbine.UI.Lotro.Button() end,
    CONTROL   = function() return Turbine.UI.Control() end,
    TEXTBOX   = function() return Turbine.UI.Lotro.TextBox() end,
    CHECKBOX  = function() return Turbine.UI.Lotro.CheckBox() end, 
    QUICKSLOT = function() return Turbine.UI.Lotro.Quickslot() end
}

UI.SCREEN_W, UI.SCREEN_H = Turbine.UI.Display.GetSize()

UI.Spacing = 4

function UI:CreateTemplate( base, template, callback )
    if self.BASE[ template ] then return end

    self.BASE[ template ] = {}

    local new_template = self.BASE[ template ]

    new_template.__index = new_template

    if base and self.BASE[ base ] then
        setmetatable( new_template, self.BASE[ base ] )
    end

    if callback then
        callback( new_template )
    end
end

function UI:Create( index, parent, callback )
    local base = self.BASE[ index ]

    if not base then return end

    local new_ui = base:CreateObject()

    if parent then
        new_ui:SetParent( parent )

        parent.child = parent.child or {}

        table.insert( parent.child, new_ui )
    else
        table.insert( self.ACTIVE, new_ui )
    end

    new_ui:SetVisible( true )

    self:CreateUtilityFunctions( new_ui )

    base:Init( new_ui )

    if callback then
        callback( new_ui, parent )
    end

    return new_ui
end