function UI:CreateUtilityFunctions( META )
    function META:SetPos( x, y )
        self:SetPosition( x, y )
    end

    function META:GetPos()
        return self:GetPosition()
    end

    function META:SetColor( red, green, blue, alpha )
        if type( red ) == "table" then
            self:SetBackColor( red )
        else
            self:SetBackColor( Turbine.UI.Color( ( alpha or 255 ) / 255, red / 255, green / 255, blue / 255 ) )
        end
    end

    function META:GetColor()
        return self:GetBackColor()
    end

    function META:Fill()
        local w, h = self:GetParent():GetSize()

        self:SetSize( w, h )
    end

    function META:Delete() --Meh

    end

    function META:GetChildren()
        return self.child or {}
    end

    function META:IsHovering()
        if not self:IsVisible() then return false end
    
        local x, y = self:GetMousePosition()
    
        local pos_x, pos_y   = self:GetPos()
        local size_x, size_y = self:GetSize()
    
        if x + pos_x >= pos_x and x + pos_x <= pos_x + size_x and y + pos_y >= pos_y and y + pos_y <= pos_y + size_y then
            self.hover = true

            return true
        else
            self.hover = false
 
            return false
        end
    end

    function META:EnableDragBase()
        self.Drag = {}
    
        self.AddDraggable = function( self, ui, draggable )
            ui.draggable = draggable or true
    
            ui.SetDraggable = function( self, bool )
                self.draggable = bool
            end
    
            table.insert( self.Drag, ui )
    
            return ui
        end
    
        self.OnDragStart = function( self, origin )
            --Override
        end
    
        self.OnDragEnd = function( self, origin, target )
            --Override
        end

        self.shift_status = false

        self:SetWantsUpdates( true )

        self.Update = function( self )
            if not self.shift_status and self:IsShiftKeyDown() then
                self.shift_status = true

                self:OnShiftChanged( true )

                self:SetMouseVisible( true )
            elseif self.shift_status and not self:IsShiftKeyDown() then
                self.shift_status = false

                self:OnShiftChanged( false )

                self:SetMouseVisible( false )
            end
        end

        self.OnShiftChanged = function( self, shift )
            --Override
        end

        self.MouseDown = function( self, mouse )
            if self.shift_status and not self.isDragging then
                for i = 1, #self.Drag do
                    local v = self.Drag[ i ]

                    if v:IsVisible() and v.draggable and v:IsHovering() then
                        local pos_x, pos_y = v:GetPos()

                        self.drag_offset_x = mouse.X - pos_x
                        self.drag_offset_y = mouse.Y - pos_y

                        self.isDragging = true

                        self.drag_origin = v

                        v.origin_x = pos_x
                        v.origin_y = pos_y

                        v.isDragging = true

                        v:SetZOrder( 10 )

                        self:OnDragStart( v )

                        break
                    end
                end
            end
        end

        self.MouseUp = function( self, mouse )
            if self.isDragging then
                self.isDragging = false
    
                local target = nil
    
                for i = 1, #self.Drag do
                    local v = self.Drag[ i ]
    
                    if not v.isDragging and v:IsVisible() and v:IsHovering() then
                        target = v

                        break
                    end
                end
    
                self.drag_origin.isDragging = false

                self.drag_origin:SetZOrder( 1 )

                self.drag_origin:SetPos( self.drag_origin.origin_x, self.drag_origin.origin_y )
    
                self:OnDragEnd( self.drag_origin, target )
            end
        end

        self.MouseMove = function( self, mouse )
            if self.isDragging then
                self.drag_origin:SetPos( mouse.X - self.drag_offset_x, mouse.Y - self.drag_offset_y )
            end
        end
    end
end
